/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimagebuffer.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGE_BUFFER_H
#define QVIMAGE_BUFFER_H

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <QObject>
#include <QRect>
#include <QDebug>
#include <ipp.h>

#include <QPoint>
#include <QSharedData>
#include <qvdefines.h>

#ifndef DOXYGEN_IGNORE_THIS

template <typename Type = uChar, int Channels = 1> class QVImageBuffer: public QSharedData
	{
	public:
		//QVImageBuffer();

		QVImageBuffer(): QSharedData() {};
		QVImageBuffer(const QVImageBuffer<Type, Channels> &);
		QVImageBuffer(uInt cols, uInt rows, uInt step = 0, const Type * buffer = NULL);

		~QVImageBuffer()	{ delete this->data; }

		uInt getRows()		const	{ return rows; }
		uInt getCols()		const	{ return cols; }
		uInt getStep()		const	{ return step; }
		uInt getChannels()	const	{ return planes; }
		uInt getTypeSize()	const	{ return typeSize; }

		int getDataSize()			const	{ return dataSize; }
		const Type *	const getReadData()	const	{ return data; }
		Type * const	getWriteData()		const	{ return data; }

	protected:
		Type * data;
		uInt cols, rows, step, planes, dataSize, typeSize;
		void allocData(uInt cols, uInt rows, uInt step = 0);
	};

template <typename Type, int Channels>
void QVImageBuffer<Type,Channels>::allocData(uInt cols, uInt rows, uInt step)
	{
	qDebug() << "QVImageBuffer<Type,Channels>::allocData(" << cols << "," << rows << "," << step << ")";
	Q_ASSERT_X(cols > 0, "QVImageBuffer::allocData()", "cols <= 0");
	Q_ASSERT_X(rows > 0, "QVImageBuffer::allocData()", "rows <= 0");
	Q_ASSERT_X((step == 0) || (step >= cols), "QVImageBuffer::allocData()", "0 < step < cols");
	this->rows = rows;
	this->cols = cols;
	this->planes = Channels;
	this->typeSize = sizeof(Type);

	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): planes" << this->planes;
	uInt temp = this->planes * this->typeSize * cols;

	if (step == 0)
		this->step = 8*(uInt)ceil((double)(temp)/8);
	else
		this->step = step;

  	Q_ASSERT_X( (this->step % 8) == 0, "QVImageBuffer::QVImageBuffer()","step % 8 != 0");
  	Q_ASSERT_X( this->step >= temp, "QVImageBuffer::QVImageBuffer()","step insufficient");

	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): rows" << this->rows;

	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): planes" << this->planes;
	this->dataSize = this->rows * this->step * this->planes;

	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): dataSize" << this->dataSize;

	if ( (this->data != NULL) && (this->cols * this->rows != cols * rows) )
		{
		delete this->data;
		this->data = NULL;
		}

	//qDebug() << "QVImageBuffer<Type,Channels>::allocData(): dataSize" << this->dataSize;

	if (this->data == NULL)
		this->data = new Type[this->dataSize];

	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): step" << this->step;
	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): step" << this->getStep();
	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): dataSize" << this->dataSize;
	qDebug() << "QVImageBuffer<Type,Channels>::allocData(): dataSize" << this->getDataSize();
	qDebug() << "QVImageBuffer<Type,Channels>::allocData(" << cols << "," << rows << "," << step << ") <- return";
	}

template <typename Type, int Channels>
QVImageBuffer<Type,Channels>::QVImageBuffer(uInt cols, uInt rows, uInt step, const Type *buffer): QSharedData(),
	data(NULL)
	{
	qDebug() << "QVImageBuffer<Type, Channels>::QVImageBuffer(" << cols << "," << rows << "," << step << ")";
	Q_ASSERT_X(cols > 0, "QVImageBuffer::allocData()", "cols <= 0");
	Q_ASSERT_X(rows > 0, "QVImageBuffer::allocData()", "rows <= 0");
	Q_ASSERT_X((step == 0) || (step >= cols), "QVImageBuffer::allocData()", "0 < step < cols");

	allocData(cols, rows, step);

	if (buffer != NULL)
		// *_* to *_* (copy)
		memcpy(this->data,buffer,this->dataSize);
	qDebug() << "QVImageBuffer<Type, Channels>::QVImageBuffer() <- return";
	}
#endif
#endif
